@@ -9,6 +9,9 @@
 #ifndef _PV_INTERNAL_H
 #define _PV_INTERNAL_H 1
 
+#ifdef HAVE_STDINT_H
+#include <stdint.h>
+#endif
 #include <stdlib.h>
 #include <signal.h>
 #include <time.h>
@@ -20,26 +23,6 @@
 extern "C" {
 #endif
 
-/*
- * Types of display component that make up an output string.
- */
-typedef enum {
-  PV_COMPONENT_STRING,		/* fixed string */
-  PV_COMPONENT_PROGRESS,	/* progress bar, with percentage if known */
-  PV_COMPONENT_BYTES,		/* number of bytes transferred */
-  PV_COMPONENT_TIMER,		/* elapsed time */
-  PV_COMPONENT_RATE,		/* current transfer rate */
-  PV_COMPONENT_AVERAGERATE,	/* average transfer rate */
-  PV_COMPONENT_ETA,		/* estimated time remaining until completion */
-  PV_COMPONENT_FINETA,		/* estimated time of completion */
-  PV_COMPONENT_NAME,		/* name prefix */
-  PV_COMPONENT_BUFPERCENT,	/* percentage of buffer used */
-  PV_COMPONENT_OUTPUTBUF,	/* recent bytes in output buffer */
-  PV_COMPONENT__MAX
-} pv_display_component;
-
-#define PV_SIZEOF_COMPONENT_STR	1024	/* size of buffer for each component */
-
 #define RATE_GRANULARITY	100000000	 /* nsec between -L rate chunks */
 #define RATE_BURST_WINDOW	5	 	 /* rate burst window (multiples of rate) */
 #define REMOTE_INTERVAL		100000000	 /* nsec between checks for -R */
@@ -55,8 +38,10 @@ typedef enum {
 
 #define PV_SIZEOF_DEFAULT_FORMAT	512
 #define PV_SIZEOF_CWD			4096
-#define PV_SIZEOF_LASTOUTPUT_BUFFER	256
+#define PV_SIZEOF_LASTWRITTEN_BUFFER	256
+#define PV_SIZEOF_PREVLINE_BUFFER	1024
 #define PV_FORMAT_ARRAY_MAX		100
+#define PV_SIZEOF_FORMAT_SEGMENTS_BUF	4096
 #define PV_SIZEOF_CRS_LOCK_FILE		1024
 
 #define PV_SIZEOF_FILE_FDINFO		4096
@@ -64,6 +49,10 @@ typedef enum {
 #define PV_SIZEOF_FILE_FDPATH		4096
 #define PV_SIZEOF_DISPLAY_NAME		512
 
+#define PV_BARSTYLE_MAX			4	/* number of different styles allowed in a format */
+#define PV_BARSTYLE_SIZEOF_STRING	10	/* max length of a bar constituent component in bytes */
+#define PV_BARSTYLE_MAX_FILLERS		10	/* max number of bar filler strings */
+
 #define PV_DISPLAY_WINDOWTITLE		1
 #define PV_DISPLAY_PROCESSTITLE		2
 
@@ -76,6 +65,72 @@ struct pvcursorstate_s {
 	bool tty_tostop_added;	/* whether any instance had to set TOSTOP on the terminal */
 };
 
+/*
+ * Types of transfer count - bytes, decimal bytes or lines.
+ */
+typedef enum {
+	PV_TRANSFERCOUNT_BYTES,
+	PV_TRANSFERCOUNT_DECBYTES,
+	PV_TRANSFERCOUNT_LINES
+} pvtransfercount_t;
+
+
+/*
+ * Structure describing a short string used as part of a progress bar, whose
+ * width in display characters may not be the same as its length in bytes.
+ */
+struct pvbarstring_spec_s {
+	char string[PV_BARSTYLE_SIZEOF_STRING];
+	uint8_t width;
+	uint8_t bytes;
+};
+
+/*
+ * Structure describing a style of progress bar.
+ *
+ * The part that moves back and forth when the size isn't known, such as
+ * "<=>", is "indicator".
+ *
+ * The string used to populate the rest of the bar is one of "filler", which
+ * contains "filler_entries" array items.  The first item "filler[0]" is
+ * used for an empty part such as " ".  The last item
+ * "filler[filler_entries-1]" is used for a 100% full part, such as "=".
+ *
+ * If there are more than the minimum of 2 entries, then the intervening
+ * entries are used for partial fills.
+ *
+ * When there are only 2 entries and the bar isn't completely empty or full,
+ * the last filled portion of the bar is filled with "tip" instead of the
+ * last filler item to indicate the tip of the bar, such as ">".
+ *
+ * Note that only the indicator is expected to ever be wider than 1 display
+ * character.  All other items are expected to have a width of 0 or 1.
+ *
+ * The "style_id" is an opaque identifier that should only be used to
+ * determine whether two styles are the same.  It is always nonzero.
+ */
+struct pvbarstyle_s {
+	uint8_t style_id;
+	uint8_t filler_entries;
+	struct pvbarstring_spec_s indicator;
+	struct pvbarstring_spec_s tip;
+	struct pvbarstring_spec_s filler[PV_BARSTYLE_MAX_FILLERS];
+};
+typedef struct pvbarstyle_s *pvbarstyle_t;
+
+
+/* Display format component type, -1 for static string. */
+typedef int8_t pvdisplay_component_t;
+#define PVDISPLAY_COMPONENT_MAX (127)	/* INT8_MAX */
+
+/* Byte count of a part of the display. */
+typedef uint16_t pvdisplay_bytecount_t;
+#define PVDISPLAY_BYTECOUNT_MAX (65535)	/* UINT16_MAX */
+
+/* Width of a part of the display. */
+typedef uint16_t pvdisplay_width_t;
+#define PVDISPLAY_WIDTH_MAX (65535)	/* UINT16_MAX */
+
 
 /*
  * Structure for holding PV internal state. Opaque outside the PV library.
@@ -112,6 +167,7 @@ struct pvstate_s {
 		/*@only@*/ /*@null@*/ char *format_string;	 /* output format string */
 		/*@only@*/ /*@null@*/ char *extra_format_string; /* extra format string */
 		/*@null@*/ char *output_name;    /* name of the output, for diagnostics */
+		/*@null@*/ char *default_bar_style; /* which bar style to use by default */
 		off_t error_skip_block;          /* skip block size, 0 for adaptive */
 		off_t rate_limit;                /* rate limit, in bytes per second */
 		size_t target_buffer_size;       /* buffer size (0=default) */
@@ -122,7 +178,7 @@ struct pvstate_s {
 		int output_fd;                   /* fd to write output to */
 		unsigned int average_rate_window; /* time window in seconds for average rate calculations */
 		unsigned int history_interval;	 /* seconds between each average rate calc history entry */
-		unsigned int width;              /* screen width */
+		pvdisplay_width_t width;         /* screen width */
 		unsigned int height;             /* screen height */
 		unsigned int extra_displays;	 /* bitmask of extra display destinations */
 		bool force;                      /* display even if not on terminal */
@@ -142,6 +198,9 @@ struct pvstate_s {
 		bool no_splice;                  /* never use splice() */
 		bool discard_input;              /* write nothing to stdout */
 		bool show_stats;		 /* show statistics on exit */
+		bool can_display_utf8;		 /* whether UTF-8 output is permitted */
+		bool can_display_colour;	 /* whether the terminal supports colour */
+		bool checked_colour_support;	 /* whether we have checked colour support yet */
 		bool width_set_manually;	 /* width was set manually, not detected */
 		bool height_set_manually;	 /* height was set manually, not detected */
 	} control;
@@ -159,13 +218,13 @@ struct pvstate_s {
 		struct sigaction old_sigint;
 		struct sigaction old_sighup;
 		struct sigaction old_sigterm;
-#ifdef SIGINFO_PROVIDES_PID
+#ifdef PV_REMOTE_CONTROL
 		struct sigaction old_sigusr2;
 #endif
 		struct sigaction old_sigalrm;
 		struct timespec tstp_time;	 /* see pv_sig_tstp() / __cont() */
 		struct timespec toffset;	 /* total time spent stopped */
-#ifdef SIGINFO_PROVIDES_PID
+#ifdef PV_REMOTE_CONTROL
 		volatile sig_atomic_t rxusr2;	 /* whether SIGUSR2 was received */
 		volatile pid_t sender;		 /* PID of sending process for SIGUSR2 */
 #endif
@@ -189,29 +248,53 @@ struct pvstate_s {
 	 *****************/
 	struct pvdisplay_s {
 
-		struct {	/* format string broken into display components */
-			size_t str_start;		/* for strings: start offset */
-			size_t str_length;		/* for strings: length */
-			pv_display_component type;	/* type of display component */
+		struct pvdisplay_segment_s {	/* format string broken into segments */
+			/* See pv__format_init() for more details. */
+			/*@dependent@*/ /*@null@*/ const char *string_parameter; /* parameter after colon in %{x:} */
+			pvdisplay_bytecount_t string_parameter_bytes;	/* number of bytes in string_parameter */
+			pvdisplay_component_t type;	/* component type, -1 for static string */
+			int8_t parameter;		/* component parameter, such as bar style index */
+			pvdisplay_width_t chosen_size;	/* "n" from %<n>A, or 0 */
+			pvdisplay_bytecount_t offset;	/* start offset of this segment in the build buffer */
+			pvdisplay_bytecount_t bytes;	/* length of segment in bytes in the build buffer */
+			pvdisplay_width_t width;	/* displayed width of segment */
 		} format[PV_FORMAT_ARRAY_MAX];
 
-		struct {	/* display components */
-			char content[PV_SIZEOF_COMPONENT_STR];	/* string to display */
-			size_t length;				/* number of bytes in string */
-			bool required;				/* true if included in format */
-		} component[PV_COMPONENT__MAX];
+		struct pvbarstyle_s barstyle[PV_BARSTYLE_MAX];
 
-		char lastoutput_buffer[PV_SIZEOF_LASTOUTPUT_BUFFER];
+		/* The last-written "n" bytes. */
+		char lastwritten_buffer[PV_SIZEOF_LASTWRITTEN_BUFFER];
+
+		/* The most recently output complete line. */
+		char previous_line[PV_SIZEOF_PREVLINE_BUFFER];
+		/* The line being received now. */
+		char next_line[PV_SIZEOF_PREVLINE_BUFFER];
 
 		/*@only@*/ /*@null@*/ char *display_buffer;	/* buffer for display string */
-		size_t display_buffer_size;	 /* size allocated to display buffer */
-		size_t display_string_len;	 /* length of string in display buffer */
-		off_t initial_offset;		 /* offset when first opened (when watching fds) */
-		size_t lastoutput_length;	 /* number of last-output bytes to show */
+		off_t initial_offset;			 /* offset when first opened (when watching fds) */
+		size_t next_line_len;				 /* length of currently receiving line so far */
 
 		size_t format_segment_count;	 /* number of format string segments */
 
-		unsigned int prev_screen_width;	 /* screen width last time we were called */
+		pvtransfercount_t count_type;	 /* type of count for transfer, rate, etc */
+
+		pvdisplay_width_t prev_screen_width;	 /* screen width last time we were called */
+
+		pvdisplay_bytecount_t display_buffer_size;	/* size allocated to display buffer */
+		pvdisplay_bytecount_t display_string_bytes;	/* byte length of string in display buffer */
+		pvdisplay_width_t display_string_width;		/* displayed width of string in display buffer */
+		pvdisplay_bytecount_t lastwritten_bytes;	 /* largest number of last-written bytes to show */
+
+		bool showing_timer;		 /* set if showing timer */
+		bool showing_bytes;		 /* set if showing byte/line count */
+		bool showing_rate;		 /* set if showing transfer rate */
+		bool showing_last_written;	 /* set if displaying the last few bytes written */
+		bool showing_previous_line;	 /* set if displaying the previously output line */
+
+		bool format_uses_colour;	 /* set if the format string uses colours */
+		bool colour_permitted;		 /* whether colour is permitted for this display */
+		bool sgr_code_active;		 /* set while SGR code is active in a display line */
+		bool final_update;		 /* set internally on the final update */
 		bool display_visible;		 /* set once anything written to terminal */
 
 	} display;
@@ -248,7 +331,7 @@ struct pvstate_s {
 
 		off_t prev_transferred;		 /* total amount transferred when called last time */
 
-		int percentage;			 /* transfer percentage completion */
+		double percentage;		 /* transfer percentage completion */
 	} calc;
 
 	/********************
@@ -306,7 +389,7 @@ struct pvstate_s {
 		off_t transferred;		 /* amount transferred (written - unconsumed) */
 
 		/* Keep track of line positions to backtrack written_but_not_consumed. */
-		/*@null@*/ off_t *line_positions; /* line separator write positions (circular buffer) */
+		/*@only@*/ /*@null@*/ off_t *line_positions; /* line separator write positions (circular buffer) */
 		size_t line_positions_capacity;	 /* total size of line position array */
 		size_t line_positions_length;	 /* number of positions stored in array */
 		size_t line_positions_head;	 /* index to use for next position */
@@ -346,6 +429,36 @@ struct pvstate_s {
 	} transfer;
 };
 
+typedef struct pvdisplay_s *pvdisplay_t;
+typedef struct pvdisplay_segment_s *pvdisplay_segment_t;
+
+/*
+ * Structure containing the parameters used by formatters.
+ */
+struct pvformatter_args_s {
+	/*@dependent@*/ pvstate_t state;		/* overall state */
+	/*@dependent@*/ pvdisplay_t display;		/* the display being updated */
+	/*@dependent@*/ pvdisplay_segment_t segment;	/* the segment of the display */
+	/*@dependent@*/ char *buffer;			/* buffer to write formatted segments into */
+	pvdisplay_bytecount_t buffer_size;		/* size of the buffer */
+	pvdisplay_bytecount_t offset;			/* current write position in the buffer */
+};
+typedef struct pvformatter_args_s *pvformatter_args_t;
+
+
+/* Pointer to a formatter function. */
+typedef pvdisplay_bytecount_t (*pvdisplay_formatter_t)(pvformatter_args_t);
+
+
+/*
+ * Structure defining a format string sequence following a %.
+ */
+struct pvdisplay_component_s {
+	/*@null@ */ const char *match;			/* string to match */
+	/*@null@ */ pvdisplay_formatter_t function;	/* function to call */
+	bool dynamic;			 /* whether it can scale with screen size */
+};
+
 
 struct pvwatchfd_s {
 #ifdef __APPLE__
@@ -366,17 +479,66 @@ struct pvwatchfd_s {
 };
 typedef struct pvwatchfd_s *pvwatchfd_t;
 
-typedef struct pvdisplay_s *pvdisplay_t;
-
 void pv_error(pvstate_t, char *, ...);
 
 int pv_main_loop(pvstate_t);
 void pv_calculate_transfer_rate(pvstate_t, bool);
+
+long pv_bound_long(long, long, long);
+long pv_seconds_remaining(const off_t, const off_t, const long double);
+void pv_si_prefix(long double *, char *, const long double, pvtransfercount_t);
+void pv_describe_amount(char *, size_t, char *, long double, char *, char *, pvtransfercount_t);
+
+int8_t pv_display_barstyle_index(pvformatter_args_t, const char *);
+
+pvdisplay_bytecount_t pv_formatter_segmentcontent(char *, pvformatter_args_t);
+
+/*
+ * Formatting functions.
+ *
+ * Each formatting function takes a state, the current display, and the
+ * segment it's for; it also takes a buffer, with a particular size, and an
+ * offset at which to start writing to the buffer.
+ *
+ * If the component is dynamically sized (such as a progress bar with no
+ * chosen_size constraint), the segment's "width" is expected to have
+ * already been populated by the caller, with the target width.
+ *
+ * The function writes the appropriate string to the buffer at the offset,
+ * and updates the segment's "offset" and "bytes".  The number of bytes
+ * written ("bytes") is also returned; it will be 0 if the string would not
+ * fit into the buffer.
+ *
+ * The caller is expected to update the segment's "width".
+ *
+ * If called with a buffer size of 0, only the side effects occur (such as
+ * setting flags like display->showing_timer).
+ */
+pvdisplay_bytecount_t pv_formatter_progress(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_progress_bar_only(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_progress_amount_only(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_bar_default(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_bar_plain(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_bar_block(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_bar_granular(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_bar_shaded(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_timer(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_eta(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_fineta(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_rate(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_average_rate(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_bytes(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_buffer_percent(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_last_written(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_previous_line(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_name(pvformatter_args_t);
+pvdisplay_bytecount_t pv_formatter_sgr(pvformatter_args_t);
+
 bool pv_format(pvstate_t, /*@null@*/ const char *, pvdisplay_t, bool, bool);
 void pv_display(pvstate_t, bool);
 ssize_t pv_transfer(pvstate_t, int, bool *, bool *, off_t, long *);
 int pv_next_file(pvstate_t, unsigned int, int);
-/*@out@*/ const char *pv_current_file_name(pvstate_t);
+/*@keep@*/ const char *pv_current_file_name(pvstate_t);
 
 void pv_write_retry(int, const char *, size_t);
 void pv_tty_write(pvstate_t, const char *, size_t);
