@@ -308,6 +308,9 @@ void display_help(void)
 		{ "-N", "--name", N_("NAME"),
 		 N_("prefix visual information with NAME"),
 		 { 0, 0, 0, 0} },
+		{ "-u", "--bar-style", N_("STYLE"),
+		 N_("set default bar style to NAME"),
+		 { 0, 0, 0, 0} },
 		{ "-x", "--extra-display", N_("SPEC"),
 		 N_("also send progress to SPEC"),
 		 { 0, 0, 0, 0} },
