@@ -19,14 +19,19 @@
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-
+#ifdef HAVE_LANGINFO_H
+#include <langinfo.h>
+#endif
 
 int pv_remote_set(opts_t, pvstate_t);
 void pv_remote_init(void);
 void pv_remote_fini(void);
 
 /*
- * Write a PID file, returning nonzero on error.
+ * Write a PID file, returning nonzero on error.  Write it atomically, such
+ * that the file either exists and contains the PID, or is not updated at
+ * all.  This is done by writing to a temporary file in the same directory
+ * first, and then renaming the temporary file to the target name.
  */
 static int pv__write_pidfile(opts_t opts)
 {
@@ -35,14 +40,27 @@ static int pv__write_pidfile(opts_t opts)
 	int pidfile_tmp_fd;
 	FILE *pidfile_tmp_fptr;
 	mode_t prev_umask;
+	const char *pidfile_template = "%s.XXXXXX";
 
 	if (NULL == opts->pidfile)
 		return 0;
 
-	pidfile_tmp_bufsize = 16 + strlen(opts->pidfile);	/* flawfinder: ignore */
 	/*
-	 * flawfinder rationale: pidfile was supplied as an argument
-	 * so we have to assume it is \0 terminated.
+	 * The buffer needs to be long enough to hold the pidfile with the
+	 * mkstemp template ".XXXXXX" after it.  The "%s" of our
+	 * pidfile_template adds 2 extra bytes to the length, of which we
+	 * need 1 byte for the terminating \0, so we subtract 1 byte more to
+	 * get the exact amount of space we need.
+	 */
+	pidfile_tmp_bufsize = strlen(pidfile_template) + strlen(opts->pidfile) - 1;	/* flawfinder: ignore */
+
+	/*
+	 * flawfinder rationale: flawfinder never likes strlen() in case
+	 * it's called on a string that isn't \0 terminated.  We have to use
+	 * strlen() to find the length of opts->pidfile, so have to trust
+	 * that the arguments in argv[] were \0 terminated.  We can be sure
+	 * that pidfile_template is \0 terminated because we've set it to a
+	 * constant value.  So we tell flawfinder to skip this check here.
 	 */
 	pidfile_tmp_name = malloc(pidfile_tmp_bufsize);
 	if (NULL == pidfile_tmp_name) {
@@ -50,7 +68,7 @@ static int pv__write_pidfile(opts_t opts)
 		return PV_ERROREXIT_REMOTE_OR_PID;
 	}
 	memset(pidfile_tmp_name, 0, pidfile_tmp_bufsize);
-	(void) pv_snprintf(pidfile_tmp_name, pidfile_tmp_bufsize, "%s.XXXXXX", opts->pidfile);
+	(void) pv_snprintf(pidfile_tmp_name, pidfile_tmp_bufsize, pidfile_template, opts->pidfile);
 
 	/*@-type@ *//* splint doesn't like mode_t */
 	prev_umask = umask(0000);	    /* flawfinder: ignore */
@@ -263,12 +281,23 @@ int main(int argc, char **argv)
 	/*@only@ */ pvstate_t state = NULL;
 	int retcode = 0;
 	bool can_have_eta = true;
+	bool can_display_utf8 = false;
+
+#if ! HAVE_SETPROCTITLE
+	initproctitle(argc, argv);
+#endif
 
 #ifdef ENABLE_NLS
 	/* Initialise language translation. */
 	(void) setlocale(LC_ALL, "");
 	(void) bindtextdomain(PACKAGE, LOCALEDIR);
 	(void) textdomain(PACKAGE);
+#ifdef HAVE_LANGINFO_H
+	/*@-mustfreefresh@ *//* splint thinks nl_langinfo() leaks memory */
+	if (0 == strcmp(nl_langinfo(CODESET), "UTF-8"))
+		can_display_utf8 = true;
+	/*@+mustfreefresh@ */
+#endif
 #endif
 
 	/* Parse the command line arguments. */
@@ -473,7 +502,9 @@ int main(int argc, char **argv)
 	pv_state_no_splice_set(state, opts->no_splice);
 	pv_state_size_set(state, opts->size);
 	pv_state_name_set(state, opts->name);
+	pv_state_default_bar_style_set(state, opts->default_bar_style);
 	pv_state_format_string_set(state, opts->format);
+	pv_state_extra_display_set(state, opts->extra_display);
 	pv_state_watch_pid_set(state, opts->watch_pid);
 	pv_state_watch_fd_set(state, opts->watch_fd);
 	pv_state_average_rate_window_set(state, opts->average_rate_window);
@@ -482,6 +513,9 @@ int main(int argc, char **argv)
 			    can_have_eta ? opts->fineta : false, opts->rate, opts->average_rate,
 			    opts->bytes, opts->bufpercent, opts->lastwritten, opts->name);
 
+	debug("%s: %s", "can_display_utf8", can_display_utf8 ? "true" : "false");
+	pv_state_set_can_display_utf8(state, can_display_utf8);
+
 	/* Initialise the signal handling. */
 	pv_sig_init(state);
 
@@ -524,5 +558,3 @@ int main(int argc, char **argv)
 
 	return retcode;
 }
-
-/* EOF */
