@@ -19,10 +19,12 @@
 #include <sys/sysmacros.h>
 #ifdef major
 #ifdef minor
+#ifdef HAVE_STRUCT_STAT_ST_RDEV
 #define CAN_BUILD_SYSFS_FILENAME 1
 #endif
 #endif
 #endif
+#endif
 #include <fcntl.h>
 #include <unistd.h>
 #ifdef HAVE_GETOPT_H
@@ -57,6 +59,8 @@ void opts_free( /*@only@ */ opts_t opts)
 		free(opts->output);
 	if (NULL != opts->store_and_forward_file)
 		free(opts->store_and_forward_file);
+	if (NULL != opts->extra_display)
+		free(opts->extra_display);
 	if (NULL != opts->argv)
 		free(opts->argv);
 	/*@+keeptrans@ */
@@ -292,7 +296,9 @@ opts_t opts_parse(unsigned int argc, char **argv)
 		{ "width", 1, NULL, (int) 'w' },
 		{ "height", 1, NULL, (int) 'H' },
 		{ "name", 1, NULL, (int) 'N' },
+		{ "bar-style", 1, NULL, (int) 'u' },
 		{ "format", 1, NULL, (int) 'F' },
+		{ "extra-display", 1, NULL, (int) 'x' },
 		{ "stats", 0, NULL, (int) 'v' },
 		{ "rate-limit", 1, NULL, (int) 'L' },
 		{ "buffer-size", 1, NULL, (int) 'B' },
@@ -317,7 +323,7 @@ opts_t opts_parse(unsigned int argc, char **argv)
 	/*@+nullassign@ */
 	int option_index = 0;
 #endif				/* HAVE_GETOPT_LONG */
-	char *short_options = "hVpteIrab8kTA:fvnqcWD:s:gl0i:w:H:N:F:L:B:CEZ:SYKXU:R:P:d:m:o:"
+	char *short_options = "hVpteIrab8kTA:fvnqcWD:s:gl0i:w:H:N:u:F:x:L:B:CEZ:SYKXU:R:P:d:m:o:"
 #ifdef ENABLE_DEBUGGING
 	    "!:"
 #endif
@@ -400,24 +406,33 @@ opts_t opts_parse(unsigned int argc, char **argv)
 				break;
 			/* falls through */
 			/*@fallthrough@ */
+		case 'L':
+			/*@fallthrough@ */
+		case 'B':
+			/*@fallthrough@ */
+		case 'Z':
+			if (!pv_getnum_check(optarg, PV_NUMTYPE_ANY_WITH_SUFFIX)) {
+				/*@-mustfreefresh@ *//* see above */
+				fprintf(stderr, "%s: -%c: %s: %s\n", opts->program_name, c, optarg,
+					_("numeric value not understood"));
+				opts_free(opts);
+				return NULL;
+				/*@+mustfreefresh@ */
+			}
+			break;
 		case 'A':
 			/*@fallthrough@ */
 		case 'w':
 			/*@fallthrough@ */
 		case 'H':
 			/*@fallthrough@ */
-		case 'L':
-			/*@fallthrough@ */
-		case 'B':
-			/*@fallthrough@ */
 		case 'R':
 			/*@fallthrough@ */
 		case 'm':
-			/*@fallthrough@ */
-		case 'Z':
-			if (!pv_getnum_check(optarg, PV_NUMTYPE_INTEGER)) {
+			if (!pv_getnum_check(optarg, PV_NUMTYPE_BARE_INTEGER)) {
 				/*@-mustfreefresh@ *//* see above */
-				fprintf(stderr, "%s: -%c: %s\n", opts->program_name, c, _("integer argument expected"));
+				fprintf(stderr, "%s: -%c: %s: %s\n", opts->program_name, c, optarg,
+					_("integer argument expected"));
 				opts_free(opts);
 				return NULL;
 				/*@+mustfreefresh@ */
@@ -426,9 +441,10 @@ opts_t opts_parse(unsigned int argc, char **argv)
 		case 'i':
 			/*@fallthrough@ */
 		case 'D':
-			if (!pv_getnum_check(optarg, PV_NUMTYPE_DOUBLE)) {
+			if (!pv_getnum_check(optarg, PV_NUMTYPE_BARE_DOUBLE)) {
 				/*@-mustfreefresh@ *//* see above */
-				fprintf(stderr, "%s: -%c: %s\n", opts->program_name, c, _("numeric argument expected"));
+				fprintf(stderr, "%s: -%c: %s: %s\n", opts->program_name, c, optarg,
+					_("numeric argument expected"));
 				opts_free(opts);
 				return NULL;
 				/*@+mustfreefresh@ */
@@ -579,6 +595,14 @@ opts_t opts_parse(unsigned int argc, char **argv)
 				return NULL;
 			}
 			break;
+		case 'u':
+			opts->default_bar_style = pv_strdup(optarg);
+			if (NULL == opts->default_bar_style) {
+				fprintf(stderr, "%s: -u: %s\n", opts->program_name, strerror(errno));
+				opts_free(opts);
+				return NULL;
+			}
+			break;
 		case 'L':
 			opts->rate_limit = pv_getnum_size(optarg, opts->decimal_units);
 			break;
@@ -617,7 +641,7 @@ opts_t opts_parse(unsigned int argc, char **argv)
 			}
 			break;
 		case 'R':
-			opts->remote = pv_getnum_count(optarg, false);
+			opts->remote = (pid_t) pv_getnum_count(optarg, false);
 			break;
 		case 'P':
 			opts->pidfile = pv_strdup(optarg);
@@ -635,6 +659,14 @@ opts_t opts_parse(unsigned int argc, char **argv)
 				return NULL;
 			}
 			break;
+		case 'x':
+			opts->extra_display = pv_strdup(optarg);
+			if (NULL == opts->extra_display) {
+				fprintf(stderr, "%s: -x: %s\n", opts->program_name, strerror(errno));
+				opts_free(opts);
+				return NULL;
+			}
+			break;
 		case 'd':
 			parse_pid = 0;
 			parse_fd = -1;
@@ -773,5 +805,3 @@ opts_t opts_parse(unsigned int argc, char **argv)
 
 	return opts;
 }
-
-/* EOF */
