@@ -17,7 +17,7 @@
  * including <ctype.h> causes weird versioned glibc dependencies on certain
  * Red Hat systems, complicating package management.
  */
-static bool pv__isdigit(char c)
+bool pv_isdigit(char c)
 {
 	return ((c >= '0') && (c <= '9')) ? true : false;
 }
@@ -37,21 +37,22 @@ off_t pv_getnum_size(const char *str, bool decimal_units)
 	unsigned int fractional_divisor = 1;
 	unsigned int binary_shift = 0;
 	off_t decimal_multiplier = 0;
+	size_t readpos = 0;
 
 	if (NULL == str)
 		return (off_t) 0;
 
 	/* Skip any non-numeric leading characters. */
-	while (str[0] != '\0' && (!pv__isdigit(str[0])))
-		str++;
+	while (str[readpos] != '\0' && (!pv_isdigit(str[readpos])))
+		readpos++;
 
 	/*
 	 * Parse the integral part of the number - the digits before the
 	 * decimal mark or units.
 	 */
-	for (; pv__isdigit(str[0]); str++) {
+	for (; pv_isdigit(str[readpos]); readpos++) {
 		integral_part = integral_part * 10;
-		integral_part += (off_t) (str[0] - '0');
+		integral_part += (off_t) (str[readpos] - '0');
 	}
 
 	/*
@@ -62,13 +63,13 @@ off_t pv_getnum_size(const char *str, bool decimal_units)
 	 * will fail if there are any locales whose decimal mark is not one
 	 * of those two characters.
 	 */
-	if (('.' == str[0]) || (',' == str[0])) {
-		str++;
-		for (; pv__isdigit(str[0]); str++) {
+	if (('.' == str[readpos]) || (',' == str[readpos])) {
+		readpos++;
+		for (; pv_isdigit(str[readpos]); readpos++) {
 			/* Stop counting below 0.0001. */
 			if (fractional_divisor < 10000) {
 				fractional_part = fractional_part * 10;
-				fractional_part += (off_t) (str[0] - '0');
+				fractional_part += (off_t) (str[readpos] - '0');
 				fractional_divisor = fractional_divisor * 10;
 			}
 		}
@@ -78,11 +79,11 @@ off_t pv_getnum_size(const char *str, bool decimal_units)
 	 * Parse any units given (K=KiB=*1024, M=MiB=1024KiB, G=GiB=1024MiB,
 	 * T=TiB=1024GiB; replace 1024 with 1000 if decimal_units is true).
 	 */
-	if (str[0] != '\0') {
+	if (str[readpos] != '\0') {
 		/* Skip any spaces or tabs after the digits. */
-		while ((' ' == str[0]) || ('\t' == str[0]))
-			str++;
-		switch (str[0]) {
+		while ((' ' == str[readpos]) || ('\t' == str[readpos]))
+			readpos++;
+		switch (str[readpos]) {
 		case 'k':
 		case 'K':
 			binary_shift = 10;
@@ -159,6 +160,8 @@ off_t pv_getnum_size(const char *str, bool decimal_units)
 	fractional_part = fractional_part / fractional_divisor;
 	integral_part += fractional_part;
 
+	debug("%s [%s] = %ld", str, decimal_units ? "decimal" : "binary", integral_part);
+
 	return integral_part;
 }
 
@@ -176,11 +179,11 @@ double pv_getnum_interval(const char *str)
 		return 0.0;
 
 	/* Skip any non-digit characters at the start. */
-	while (str[0] != '\0' && (!pv__isdigit(str[0])))
+	while (str[0] != '\0' && (!pv_isdigit(str[0])))
 		str++;
 
 	/* Parse the digits before the decimal mark. */
-	for (; pv__isdigit(str[0]); str++) {
+	for (; pv_isdigit(str[0]); str++) {
 		result = result * 10;
 		result += (double) (str[0] - '0');
 	}
@@ -193,7 +196,7 @@ double pv_getnum_interval(const char *str)
 	str++;
 
 	/* Parse the digits after the decimal mark, up to 0.0000001. */
-	for (; pv__isdigit(str[0]) && step < 1000000; str++) {
+	for (; pv_isdigit(str[0]) && step < 1000000; str++) {
 		step = step * 10;
 		result += ((double) (str[0] - '0')) / step;
 	}
@@ -226,31 +229,31 @@ bool pv_getnum_check(const char *str, pv_numtype type)
 		str++;
 
 	/* If the next character isn't a digit, this isn't a number. */
-	if (!pv__isdigit(str[0]))
+	if (!pv_isdigit(str[0]))
 		return false;
 
 	/* Skip over the digits. */
-	for (; pv__isdigit(str[0]); str++);
+	for (; pv_isdigit(str[0]); str++);
 
 	/*
 	 * If there's a decimal mark (see note in pv_getnum_size() above),
 	 * check that too.
 	 */
 	if (('.' == str[0]) || (',' == str[0])) {
-		/* Integers should have no decimal mark. */
-		if (type == PV_NUMTYPE_INTEGER)
+		/* Bare integers should have no decimal mark. */
+		if (type == PV_NUMTYPE_BARE_INTEGER)
 			return false;
 		/* Skip the decimal mark, then all digits. */
 		str++;
-		for (; pv__isdigit(str[0]); str++);
+		for (; pv_isdigit(str[0]); str++);
 	}
 
 	/* If the string ends here, this is a valid number. */
 	if ('\0' == str[0])
 		return true;
 
-	/* A units suffix is not allowed for doubles, only for integers. */
-	if (type == PV_NUMTYPE_DOUBLE)
+	/* A units suffix is only allowed for ANY_WITH_SUFFIX. */
+	if (type != PV_NUMTYPE_ANY_WITH_SUFFIX)
 		return false;
 
 	/* Skip trailing spaces or tabs. */
@@ -282,16 +285,20 @@ bool pv_getnum_check(const char *str, pv_numtype type)
 
 
 /*
- * Return an integer representing "amount" as a percentage of "total", i.e.
+ * Return a value representing "amount" as a percentage of "total", i.e.
  * 100*amount/total.  If "total" is zero or less, return 0.
  */
-int pv_percentage(off_t amount, const off_t total)
+double pv_percentage(off_t amount, const off_t total)
 {
+	double amount_as_double;
+
 	if (total < 1)
-		return 0;
+		return 0.0;
 
-	amount *= 100;
-	amount /= total;
+	amount_as_double = (double) amount;
 
-	return (int) amount;
+	amount_as_double *= 100.0;
+	amount_as_double /= (double) total;
+
+	return amount_as_double;
 }
