@@ -1,7 +1,7 @@
 /*
  * State management functions.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -10,22 +10,26 @@
 #include "pv.h"
 #include "pv-internal.h"
 
+/* We do not set this because it breaks "dd" - see below. */
+/* #undef MAKE_OUTPUT_NONBLOCKING */
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 #include <errno.h>
+#include <fcntl.h>
 
 
 /* alloc / realloc history buffer */
 static void pv_alloc_history(pvstate_t state)
 {
-	if (NULL != state->display.history)
-		free(state->display.history);
-	state->display.history = NULL;
+	if (NULL != state->calc.history)
+		free(state->calc.history);
+	state->calc.history = NULL;
 
-	state->display.history = calloc((size_t) (state->display.history_len), sizeof(state->display.history[0]));
-	if (NULL == state->display.history) {
+	state->calc.history = calloc((size_t) (state->calc.history_len), sizeof(state->calc.history[0]));
+	if (NULL == state->calc.history) {
 		/*@-mustfreefresh@ */
 		/*
 		 * splint note: the gettext calls made by _() cause memory
@@ -38,10 +42,69 @@ static void pv_alloc_history(pvstate_t state)
 		return;
 	}
 
-	state->display.history_first = state->display.history_last = 0;
-	state->display.history[0].elapsed_sec = 0.0;	/* to be safe, memset() not recommended for doubles */
+	state->calc.history_first = state->calc.history_last = 0;
+	state->calc.history[0].elapsed_sec = 0.0;	/* to be safe, memset() not recommended for doubles */
 }
 
+
+/*
+ * Clear the calculated parts of a state structure.
+ */
+void pv_state_reset(pvstate_t state)
+{
+	if (NULL == state)
+		return;
+
+	state->flag.reparse_display = 1;
+	state->status.current_input_file = -1;
+
+	state->display.initial_offset = 0;
+	state->display.display_visible = false;
+
+	state->extra_display.initial_offset = 0;
+	state->extra_display.display_visible = false;
+
+	/*
+	 * Explicitly set important floating point values to 0, as memset()
+	 * is not recommended for this.
+	 */
+	state->calc.transfer_rate = 0.0;
+	state->calc.average_rate = 0.0;
+	state->calc.prev_elapsed_sec = 0.0;
+	state->calc.prev_rate = 0.0;
+	state->calc.prev_trans = 0.0;
+	state->calc.current_avg_rate = 0.0;
+	state->calc.rate_min = 0.0;
+	state->calc.rate_max = 0.0;
+	state->calc.rate_sum = 0.0;
+	state->calc.ratesquared_sum = 0.0;
+	state->calc.measurements_taken = 0;
+	state->calc.prev_transferred = 0;
+	state->calc.percentage = 0.0;
+	state->calc.history_first = state->calc.history_last = 0;
+	if (NULL != state->calc.history) {
+		state->calc.history[0].elapsed_sec = 0.0;
+	}
+
+	state->transfer.elapsed_seconds = 0.0;
+	state->transfer.read_position = 0;
+	state->transfer.write_position = 0;
+	state->transfer.to_write = 0;
+	state->transfer.written = 0;
+	state->transfer.total_written = 0;
+	state->transfer.written_but_not_consumed = 0;
+	state->transfer.read_errors_in_a_row = 0;
+	state->transfer.last_read_skip_fd = 0;
+#ifdef HAVE_SPLICE
+	state->transfer.splice_failed_fd = -1;
+#endif				/* HAVE_SPLICE */
+
+	state->transfer.line_positions_length = 0;
+	state->transfer.line_positions_head = 0;
+	state->transfer.last_output_position = 0;
+}
+
+
 /*
  * Create a new state structure, and return it, or 0 (NULL) on error.
  */
@@ -67,18 +130,14 @@ pvstate_t pv_state_alloc(const char *program_name)
 
 	state->control.watch_pid = 0;
 	state->control.watch_fd = -1;
+	state->control.output_fd = -1;
 #ifdef HAVE_IPC
 	state->cursor.shmid = -1;
 	state->cursor.pvcount = 1;
 #endif				/* HAVE_IPC */
 	state->cursor.lock_fd = -1;
 
-	state->flag.reparse_display = 1;
-	state->status.current_input_file = -1;
-#ifdef HAVE_SPLICE
-	state->transfer.splice_failed_fd = -1;
-#endif				/* HAVE_SPLICE */
-	state->display.display_visible = false;
+	pv_state_reset(state);
 
 	/*
 	 * Get the current working directory, if possible, as a base for
@@ -106,6 +165,26 @@ void pv_state_free(pvstate_t state)
 	if (0 == state)
 		return;
 
+	/*
+	 * Close the output file first, so we can report any errors while we
+	 * still know the program name and output filename.
+	 */
+	if (state->control.output_fd >= 0) {
+		if (STDOUT_FILENO != state->control.output_fd) {
+			if (close(state->control.output_fd) < 0) {
+				fprintf(stderr, "%s: %s: %s\n", state->status.program_name,
+					NULL == state->control.output_name ? "(null)" : state->control.output_name,
+					strerror(errno));
+			}
+		}
+		state->control.output_fd = -1;
+	}
+
+	if (NULL != state->control.output_name) {
+		free(state->control.output_name);
+		state->control.output_name = NULL;
+	}
+
 	if (NULL != state->status.program_name)
 		free(state->status.program_name);
 	state->status.program_name = NULL;
@@ -114,16 +193,30 @@ void pv_state_free(pvstate_t state)
 		free(state->display.display_buffer);
 	state->display.display_buffer = NULL;
 
+	if (NULL != state->extra_display.display_buffer)
+		free(state->extra_display.display_buffer);
+	state->extra_display.display_buffer = NULL;
+
 	if (NULL != state->control.name) {
 		free(state->control.name);
 		state->control.name = NULL;
 	}
 
+	if (NULL != state->control.default_bar_style) {
+		free(state->control.default_bar_style);
+		state->control.default_bar_style = NULL;
+	}
+
 	if (NULL != state->control.format_string) {
 		free(state->control.format_string);
 		state->control.format_string = NULL;
 	}
 
+	if (NULL != state->control.extra_format_string) {
+		free(state->control.extra_format_string);
+		state->control.extra_format_string = NULL;
+	}
+
 	/*@-keeptrans@ */
 	if (NULL != state->transfer.transfer_buffer)
 		free(state->transfer.transfer_buffer);
@@ -131,9 +224,13 @@ void pv_state_free(pvstate_t state)
 	/*@+keeptrans@ */
 	/* splint - explicitly freeing this structure, so free() here is OK. */
 
-	if (NULL != state->display.history)
-		free(state->display.history);
-	state->display.history = NULL;
+	if (NULL != state->transfer.line_positions)
+		free(state->transfer.line_positions);
+	state->transfer.line_positions = NULL;
+
+	if (NULL != state->calc.history)
+		free(state->calc.history);
+	state->calc.history = NULL;
 
 	if (NULL != state->files.filename) {
 		unsigned int file_idx;
@@ -209,6 +306,11 @@ void pv_state_cursor_set(pvstate_t state, bool val)
 	state->control.cursor = val;
 }
 
+void pv_state_show_stats_set(pvstate_t state, bool val)
+{
+	state->control.show_stats = val;
+}
+
 void pv_state_numeric_set(pvstate_t state, bool val)
 {
 	state->control.numeric = val;
@@ -224,6 +326,11 @@ void pv_state_delay_start_set(pvstate_t state, double val)
 	state->control.delay_start = val;
 }
 
+void pv_state_rate_gauge_set(pvstate_t state, bool val)
+{
+	state->control.rate_gauge = val;
+}
+
 void pv_state_linemode_set(pvstate_t state, bool val)
 {
 	state->control.linemode = val;
@@ -234,6 +341,11 @@ void pv_state_bits_set(pvstate_t state, bool bits)
 	state->control.bits = bits;
 }
 
+void pv_state_decimal_units_set(pvstate_t state, bool decimal_units)
+{
+	state->control.decimal_units = decimal_units;
+}
+
 void pv_state_null_terminated_lines_set(pvstate_t state, bool val)
 {
 	state->control.null_terminated_lines = val;
@@ -302,7 +414,9 @@ void pv_state_interval_set(pvstate_t state, double val)
 
 void pv_state_width_set(pvstate_t state, unsigned int val, bool was_set_manually)
 {
-	state->control.width = val;
+	if (val > PVDISPLAY_WIDTH_MAX)
+		val = PVDISPLAY_WIDTH_MAX;
+	state->control.width = (pvdisplay_width_t) val;
 	state->control.width_set_manually = was_set_manually;
 }
 
@@ -322,6 +436,16 @@ void pv_state_name_set(pvstate_t state, /*@null@ */ const char *val)
 		state->control.name = pv_strdup(val);
 }
 
+void pv_state_default_bar_style_set(pvstate_t state, /*@null@ */ const char *val)
+{
+	if (NULL != state->control.default_bar_style) {
+		free(state->control.default_bar_style);
+		state->control.default_bar_style = NULL;
+	}
+	if (NULL != val)
+		state->control.default_bar_style = pv_strdup(val);
+}
+
 void pv_state_format_string_set(pvstate_t state, /*@null@ */ const char *val)
 {
 	if (NULL != state->control.format_string) {
@@ -332,6 +456,57 @@ void pv_state_format_string_set(pvstate_t state, /*@null@ */ const char *val)
 		state->control.format_string = pv_strdup(val);
 }
 
+void pv_state_extra_display_set(pvstate_t state, /*@null@ */ const char *val)
+{
+	const char *word_start;
+	size_t offset;
+
+	if (NULL != state->control.extra_format_string) {
+		free(state->control.extra_format_string);
+		state->control.extra_format_string = NULL;
+	}
+
+	state->control.extra_displays = 0;
+	if (NULL == val)
+		return;
+
+	word_start = val;
+	while (NULL != word_start && '\0' != word_start[0]) {
+		offset = 0;
+		while ('\0' != word_start[offset] && ',' != word_start[offset] && ':' != word_start[offset])
+			offset++;
+		if (((11 == offset) && (0 == strncmp(word_start, "windowtitle", 11)))
+		    || ((6 == offset) && (0 == strncmp(word_start, "window", 6)))
+		    ) {
+			debug("%s", "enabling windowtitle");
+			state->control.extra_displays |= PV_DISPLAY_WINDOWTITLE;
+		} else if (((12 == offset) && (0 == strncmp(word_start, "processtitle", 12)))
+			   || ((9 == offset) && (0 == strncmp(word_start, "proctitle", 9)))
+			   || ((7 == offset) && (0 == strncmp(word_start, "process", 7)))
+			   || ((4 == offset) && (0 == strncmp(word_start, "proc", 4)))
+		    ) {
+			debug("%s", "enabling processtitle");
+			state->control.extra_displays |= PV_DISPLAY_PROCESSTITLE;
+		}
+		switch (word_start[offset]) {
+		case ',':
+			offset++;
+			break;
+		case ':':
+			offset++;
+			debug("%s: [%s]", "setting extra_format_string", word_start + offset);
+			state->control.extra_format_string = pv_strdup(word_start + offset);
+			word_start = NULL;
+			break;
+		default:
+			word_start = NULL;
+			break;
+		}
+		if (NULL != word_start)
+			word_start += offset;
+	}
+}
+
 void pv_state_watch_pid_set(pvstate_t state, pid_t val)
 {
 	state->control.watch_pid = val;
@@ -342,21 +517,53 @@ void pv_state_watch_fd_set(pvstate_t state, int val)
 	state->control.watch_fd = val;
 }
 
+void pv_state_output_set(pvstate_t state, int fd, const char *name)
+{
+	/*
+	 * Close any previous output file first, so we can report any errors
+	 * before we store the new output filename.
+	 */
+	if (state->control.output_fd >= 0 && state->control.output_fd != STDOUT_FILENO) {
+		if (close(state->control.output_fd) < 0) {
+			fprintf(stderr, "%s: %s: %s\n", state->status.program_name,
+				NULL == state->control.output_name ? "(null)" : state->control.output_name,
+				strerror(errno));
+		}
+	}
+	if (NULL != state->control.output_name)
+		free(state->control.output_name);
+	state->control.output_fd = fd;
+	state->control.output_name = pv_strdup(name);
+#ifdef MAKE_OUTPUT_NONBLOCKING
+	/*
+	 * Try and make the output use non-blocking I/O.
+	 *
+	 * Note that this can cause problems with (broken) applications
+	 * such as dd when used in a pipeline.
+	 */
+	fcntl(state->control.output_fd, F_SETFL, O_NONBLOCK | fcntl(state->control.output_fd, F_GETFL));
+#endif				/* MAKE_OUTPUT_NONBLOCKING */
+}
+
 void pv_state_average_rate_window_set(pvstate_t state, unsigned int val)
 {
 	if (val < 1)
 		val = 1;
 	state->control.average_rate_window = val;
 	if (val >= 20) {
-		state->display.history_len = (size_t) (val / 5 + 1);
-		state->display.history_interval = 5;
+		state->calc.history_len = (size_t) (val / 5 + 1);
+		state->control.history_interval = 5;
 	} else {
-		state->display.history_len = (size_t) (val + 1);
-		state->display.history_interval = 1;
+		state->calc.history_len = (size_t) (val + 1);
+		state->control.history_interval = 1;
 	}
 	pv_alloc_history(state);
 }
 
+void pv_state_set_can_display_utf8(pvstate_t state, bool val)
+{
+	state->control.can_display_utf8 = val;
+}
 
 /*
  * Set the array of input files.
@@ -404,5 +611,3 @@ void pv_state_inputfiles(pvstate_t state, unsigned int input_file_count, const c
 
 /*@+nullstate@*/
 /* splint: see unqualifiedtrans note by free() above. */
-
-/* EOF */
