@@ -113,7 +113,7 @@ static int is_data_ready(int fd_in, /*@null@ */ bool *fd_in_ready, int fd_out, /
  * We stop retrying if the time elapsed since this function was entered
  * reaches TRANSFER_READ_TIMEOUT seconds.
  */
-static ssize_t pv__transfer_read_repeated(int fd, void *buf, size_t count)
+static ssize_t pv__transfer_read_repeated(int fd, char *buf, size_t count)
 {
 	struct timespec start_time;
 	ssize_t total_read;
@@ -193,7 +193,7 @@ static ssize_t pv__transfer_read_repeated(int fd, void *buf, size_t count)
  * We stop retrying if the time elapsed since this function was entered
  * reaches TRANSFER_WRITE_TIMEOUT seconds.
  */
-static ssize_t pv__transfer_write_repeated(int fd, void *buf, size_t count, bool sync_after_write)
+static ssize_t pv__transfer_write_repeated(int fd, char *buf, size_t count, bool sync_after_write)
 {
 	struct timespec start_time;
 	ssize_t total_written;
@@ -460,15 +460,7 @@ static int pv__transfer_read(pvstate_t state, int fd, bool *eof_in, bool *eof_ou
 	 * reached the end of this file.
 	 */
 	if (do_not_skip_errors) {
-		/*@-compdef@ */
 		pv_error(state, "%s: %s: %s", pv_current_file_name(state), _("read failed"), strerror(errno));
-		/*@+compdef@ */
-		/*
-		 * splint says the storage pointed to by the result of
-		 * pv_current_file_name() is not fully defined.
-		 *
-		 * TODO: investigate and fix the reason for this.
-		 */
 		*eof_in = true;
 		if (state->transfer.write_position >= state->transfer.read_position) {
 			*eof_out = true;
@@ -718,28 +710,42 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 	}
 
 	if (nwritten > 0) {
+		bool tracking_lines = false;
+
+		if ((state->control.linemode) && (lineswritten != NULL))
+			tracking_lines = true;
+		else if (state->display.showing_previous_line)
+			tracking_lines = true;
+
 		/*
 		 * Write returned >0 - data successfully written.
 		 */
-		if ((state->control.linemode) && (lineswritten != NULL)) {
+		if (tracking_lines) {
 			char separator;
 			char *ptr;
 			long lines = 0;
 
 			/*
-			 * Line mode - look through what we've just written
-			 * to count how many lines there were.
+			 * Tracking lines - either line mode, or we're
+			 * showing the last line in the display, or both.
+			 * So we need to look through what we've just
+			 * written to either count how many lines there
+			 * were, or get the content of the most recent
+			 * complete line, or both.
 			 */
 
 			/* Allocate buffer to remember line positions. */
-			if (NULL == state->transfer.line_positions) {
+			if (NULL == state->transfer.line_positions && NULL != lineswritten) {
 				state->transfer.line_positions_capacity = MAX_LINE_POSITIONS;
+				/*@-mustfreeonly@ */
 				state->transfer.line_positions =
 				    calloc((size_t) (state->transfer.line_positions_capacity), sizeof(off_t));
 				if (NULL == state->transfer.line_positions) {
 					pv_error(state, "%s: %s", _("line position buffer allocation failed"),
 						 strerror(errno));
 				}
+				/*@+mustfreeonly@ */
+				/* splint doesn't see we only call calloc() when line_positions is NULL. */
 			}
 
 			if (state->control.null_terminated_lines) {
@@ -752,12 +758,50 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 			for (ptr++;
 			     ptr - (char *) state->transfer.transfer_buffer - state->transfer.write_position <
 			     (size_t) nwritten; ptr++, state->transfer.last_output_position++) {
-				if (*ptr != separator)
+				if (*ptr != separator) {
+					/*
+					 * If we're displaying the previous
+					 * line ("%L"), add to our line
+					 * buffer.
+					 */
+					if (state->display.showing_previous_line
+					    && state->display.next_line_len < PV_SIZEOF_PREVLINE_BUFFER - 1) {
+						state->display.next_line[state->display.next_line_len] = *ptr;
+						state->display.next_line_len++;
+					}
 					continue;
+				}
 
 				/* Separator found - increment line count. */
 				++lines;
 
+				/*
+				 * If we're displaying the previous line
+				 * ("%L"), update the previous-line buffer
+				 * with the line we just completed, and
+				 * start a new one.
+				 */
+				if (state->display.showing_previous_line) {
+					memset(state->display.previous_line, 0, PV_SIZEOF_PREVLINE_BUFFER);
+					if (state->display.next_line_len > PV_SIZEOF_PREVLINE_BUFFER - 1)
+						state->display.next_line_len = PV_SIZEOF_PREVLINE_BUFFER - 1;
+					if (state->display.next_line_len > 0) {
+						memcpy(state->display.previous_line, state->display.next_line,	/* flawfinder: ignore */
+						       state->display.next_line_len);
+						debug("%s: [%s]", "updated previous_line",
+						      state->display.previous_line);
+					}
+					state->display.next_line_len = 0;
+					/*
+					 * flawfinder - next_line_len is
+					 * guaranteed to be less than the
+					 * size of the previous_line buffer
+					 * since we check it just before
+					 * memcpy(), and we ensure that the
+					 * last byte in the buffer is \0.
+					 */
+				}
+
 				if (NULL == state->transfer.line_positions)
 					continue;
 
@@ -777,7 +821,8 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 				}
 			}
 
-			*lineswritten += lines;
+			if (NULL != lineswritten)
+				*lineswritten += lines;
 		}
 
 		state->transfer.write_position += nwritten;
@@ -787,39 +832,39 @@ static int pv__transfer_write(pvstate_t state, bool *eof_in, bool *eof_out, long
 		 * If we're monitoring the output, update our copy of the
 		 * last few bytes we've written.
 		 */
-		if (state->display.component[PV_COMPONENT_OUTPUTBUF].required && (nwritten > 0)) {
-			size_t new_portion_length, old_portion_length;
+		if (state->display.showing_last_written && (nwritten > 0)) {
+			size_t new_portion_size, old_portion_size;
 
-			new_portion_length = (size_t) nwritten;
-			if (new_portion_length > state->display.lastoutput_length)
-				new_portion_length = state->display.lastoutput_length;
+			new_portion_size = (size_t) nwritten;
+			if (new_portion_size > state->display.lastwritten_bytes)
+				new_portion_size = state->display.lastwritten_bytes;
 
-			old_portion_length = state->display.lastoutput_length - new_portion_length;
+			old_portion_size = state->display.lastwritten_bytes - new_portion_size;
 
 			/*
 			 * Make room for the new portion.
 			 */
-			if (old_portion_length > 0) {
-				memmove(state->display.lastoutput_buffer,
-					state->display.lastoutput_buffer + new_portion_length, old_portion_length);
+			if (old_portion_size > 0) {
+				memmove(state->display.lastwritten_buffer,
+					state->display.lastwritten_buffer + new_portion_size, old_portion_size);
 			}
 
 			/*
 			 * Copy the new data in.
 			 */
-			memcpy(state->display.lastoutput_buffer +	/* flawfinder: ignore */
-			       old_portion_length,
-			       state->transfer.transfer_buffer + state->transfer.write_position - new_portion_length,
-			       new_portion_length);
+			memcpy(state->display.lastwritten_buffer +	/* flawfinder: ignore */
+			       old_portion_size,
+			       state->transfer.transfer_buffer + state->transfer.write_position - new_portion_size,
+			       new_portion_size);
 			/*
 			 * flawfinder rationale: calculations above ensure
-			 * that old_portion_length + new_portion_length is
-			 * always <= lastoutput_length, and
-			 * lastoutput_length is guaranteed by
+			 * that old_portion_size + new_portion_size is
+			 * always <= lastwritten_bytes, and
+			 * lastwritten_bytes is guaranteed by
 			 * pv__format_init() to be no more than
-			 * PV_SIZEOF_LASTOUTPUT_BUFFER, which is the size of
-			 * lastoutput_buffer, so the memcpy() will always
-			 * fit into the buffer.
+			 * PV_SIZEOF_LASTWRITTEN_BUFFER, which is the size
+			 * of lastwritten_buffer, so the memcpy() will
+			 * always fit into the buffer.
 			 */
 		}
 
