@@ -308,6 +308,7 @@ static int pidfds(pvstate_t state, unsigned int pid, struct proc_fdinfo **fds, i
 		pv_error(state, "%s: alloc failed: %s", _("pid"), strerror(errno));
 		return -1;
 	}
+	memset(*fds, 0, size_needed);
 
 	proc_pidinfo(pid, PROC_PIDLISTFDS, 0, *fds, size_needed);
 
@@ -419,9 +420,12 @@ int pv_watchpid_scanfds(pvstate_t state,
 #else
 		if (sscanf(d->d_name, "%d", &fd) != 1)
 			continue;
+#endif
+
+		/* Skip if the fd is outside the array. */
 		if ((fd < 0) || (fd >= FD_SETSIZE))
 			continue;
-#endif
+
 		/*
 		 * Skip if this fd is already known to us.
 		 */
